;**********************************************************************
;                                                                     *
;    Filename:	    usbpw.asm                                         *
;    Date:          2006.03.01.                                       *
;    File Version:  1.2                                               *
;                                                                     *
;    Author:        M.Urushidani                                      *
;    Company:                                                         *
;**********************************************************************
;    Files required:   p16f873.inc                                    *
;**********************************************************************
;    Notes:     Use Assembler 'MPASM'                                 *
;**********************************************************************

	list		p=16f873	; list directive to define processor
	#include	<p16f873.inc>	; processor specific variable definitions
	
	__CONFIG _CP_OFF & _WDT_OFF & _BODEN_OFF & _PWRTE_ON & _HS_OSC & _WRT_ENABLE_OFF & _LVE_OFF

;********************** VARIABLE DEFINITIONS***************************
counter		EQU	0x20		;A/D wait counter
pos_ad		EQU	0x21
pos_tgt		EQU	0x22
pos_duty	EQU	0x23
pos_err		EQU	0x26
w_temp		EQU	0x27		;interrupt push
status_temp	EQU	0x28		;interrupt push
neg_ad		EQU	0x2A
neg_tgt		EQU	0x2B
neg_duty	EQU	0x2C
neg_err		EQU	0x2D
work1		EQU	0x2E		;work in main
work2		EQU	0x2F
work3		EQU	0x30
count		EQU	work1
chrdata		EQU	work2
tbldata		EQU	work3
_data		EQU	0x38		;for table jump
_C		EQU	0x40
_D		EQU	0x41
_E		EQU	0x42
ptgtvolt	EQU	0x43		;main/int interface
pos_curr	EQU	0x44
neg_curr	EQU	0x45
ntgtvolt	EQU	0x46
;**********************************************************************
	ORG     0x000           	; processor reset vector
	GOTO	start
;**********************************************************************
	ORG	0x004			;interrupt vector
	GOTO	tm0int
;**************************** ROM TABLE *******************************
text_snum
	ADDWF	PCL,f
	dt	"ACKSN000000000",0x0d,0
text_pnum
	ADDWF	PCL,f
	dt	"ACKPNUSBPWRV01",0x0d,0
text_acka
	ADDWF	PCL,f
	dt	"ACKA",0

;**************************** interrupt routine ***********************
tm0int	MOVWF	w_temp			;push W
	SWAPF	STATUS,W		;push STATUS
	MOVWF	status_temp

					;******** A/Ds **********
	MOVLW	0x81			;ADon,f/32,AN0
	MOVWF	ADCON0
	CALL	time15			;wait acquisition time
	BSF	ADCON0,GO		;ADstart
wait0
	BTFSC	ADCON0,GO
	GOTO	wait0
	MOVF	ADRESH,W		;get AD data
	movwf	pos_ad			;W->pos_ad

	MOVLW	0x89			;10001001,ADon,f/32,AN1
	MOVWF	ADCON0
	CALL	time15			;wait acquisition time
	BSF	ADCON0,GO		;ADstart
wait0n
	BTFSC	ADCON0,GO
	GOTO	wait0n
	MOVF	ADRESH,W		;get AD data
	movwf	neg_ad			;W->neg_ad

	MOVLW	0x91			;ADon,f/32,AN2
	MOVWF	ADCON0
	CALL	time15			;wait acquisition time
	BSF	ADCON0,GO		;ADstart
waitan2
	BTFSC	ADCON0,GO
	GOTO	waitan2
	MOVF	ADRESH,W		;get AD data
	movwf	pos_curr		;W->pos_curr

	MOVLW	0x99			;ADon,f/32,AN3
	MOVWF	ADCON0
	CALL	time15			;wait acquisition time
	BSF	ADCON0,GO		;ADstart
waitan3
	BTFSC	ADCON0,GO
	GOTO	waitan3
	MOVF	ADRESH,W		;get AD data
	movwf	neg_curr		;W->neg_curr

					;******** positive pwm *******

	MOVF	CCPR1L,W		;get duty
	MOVWF	pos_duty		;

	movf	pos_tgt,W		;get target
	xorwf	pos_ad,W		;W xor pos_ad ->W
	btfsc	STATUS,Z
	goto	duty_end		;pos_tgt==pos_ad
	movf	pos_tgt,W		;pos_tgt ->W
	subwf	pos_ad,W		;pos_ad - pos_tgt ->W
	btfss	STATUS,C		;bit test STATUS.C, skip if set
	goto	up_duty			;ad<tgt
	decf	pos_duty,W		;ad>tgt
	movwf	pos_duty
	xorlw	0xff
	btfss	STATUS,Z
	goto	not_pbtm
	movlw	0x00			;lower limit
	movwf	pos_duty
not_pbtm
	GOTO	duty_end
up_duty
	incf	pos_duty,W
	movwf	pos_duty
	xorlw	0x00
	btfss	STATUS,Z
	goto	duty_end
	movlw	0xff
	movwf	pos_duty
duty_end
	movf	pos_duty,W		;pos_duty->W
	MOVWF	CCPR1L			;set PWM duty

					;******* negative pwm *******

	MOVF	CCPR2L,W		;get duty
	MOVWF	neg_duty		;
	movf	neg_tgt,W		;get target
	xorwf	neg_ad,W		;W xor neg_ad ->W
	btfsc	STATUS,Z
	goto	nduty_end		;neg_tgt==neg_ad
	movf	neg_tgt,W		;neg_tgt ->W
	subwf	neg_ad,W		;neg_ad - neg_tgt ->W
	btfss	STATUS,C		;skip if C=1(not borrow)
	goto	up_nduty		;ad<tgt
	decf	neg_duty,W		;ad>tgt
	movwf	neg_duty
	xorlw	0xff
	btfss	STATUS,Z
	goto	not_nbtm
	movlw	0x00			;lower limit
	movwf	neg_duty
not_nbtm	
	GOTO	nduty_end
up_nduty
	incf	neg_duty,W
	movwf	neg_duty
	xorlw	0x00
	btfss	STATUS,Z
	goto	nduty_end
	movlw	0xff
	movwf	neg_duty

nduty_end
	movf	neg_duty,W		;neg_duty->W
	MOVWF	CCPR2L			;set PWM duty

term
	BCF	INTCON,T0IF		;clear interrupt flag

	SWAPF	status_temp,W		;pop STATUS
	MOVWF	STATUS
	SWAPF	w_temp,F		;pop W
	SWAPF	w_temp,W
	
	RETFIE
;************************ subroutine in interrupt *********************

time15
	MOVLW	0Dh			;wait15usec
	MOVWF	counter
t_lp1	DECFSZ	counter,F		;2+3*12-1=37
	GOTO	t_lp1
	RETURN				;=38*0.4us

;***************************** reset start ****************************

start
;*** Port setting ***
	BSF	STATUS,RP0		;bank1
	MOVLW	0xff
	MOVWF	TRISB			;RB=in
	MOVLW	0x30			;00110000
	MOVWF	TRISC
;*** Timer0 setting ***
	MOVLW	0x40			;drift<->ripple
	MOVWF	OPTION_REG		;bit7:weak pull-up
;*** PWM setting ***
	MOVLW	0xFF			;period
	MOVWF	PR2
	BCF	STATUS,RP0		;bank0
;*** PWM mode ***
	MOVLW	0x0c			;PWM mode
	MOVWF	CCP1CON			;pos PWM
	MOVWF	CCP2CON			;neg PWM
;*** PWM duty ***
	MOVLW	0x00			;duty 0%
	MOVWF	CCPR1L			;pos duty
	MOVWF	CCPR2L			;neg duty
;*** Timer2 set and start ***
	MOVLW	0x04			;Timer2 prescale=1
	MOVWF	T2CON			;Timer2 start
;*** AD set ***
	BCF	STATUS,RP0		;bank0
	CLRF	PORTA			;initialize data latch
	BSF	STATUS,RP0		;bank1
	MOVLW	0x02			;AN0...AN4
	MOVWF	ADCON1			;set AD function
	MOVLW	0xCF			;11001111
	MOVWF	TRISA			;portA direction
	BCF	STATUS,RP0		;bank0
	MOVLW	0x80			;f/32,ch0,ADoff
	MOVWF	ADCON0
;*** PortC initial data
	MOVLW	0x81			;10000001
	MOVWF	PORTC
	clrf	pos_tgt
	clrf	neg_tgt
;*** Interrupt start ***
	BSF	INTCON,T0IE		;enable timer0 int
	BSF	INTCON,GIE		;GIE on

;****************************** main **********************************

main					;communicate with host
	BSF	PORTC,7			;LED OFF
	BSF	INTCON,GIE		;EI
	CLRF	count			;set counter to zero
waitcmd					;wait for command
	BTFSC	PORTC,5			;RXF#=L?
	GOTO	waitcmd
	BCF	PORTC,7			;command in,LED-ON
	BCF	INTCON,GIE		;DI
	BCF	PORTC,0			;RD#->L
	MOVF	PORTB,W			;portB->W
	BSF	PORTC,0			;RD#->H
	MOVWF	chrdata			;W->chrdata
;-------------------------------	;chrdata -> RAM
	movlw	_data			;_data top address ->W
	addwf	count,W			;_data + count ->W
	movwf	FSR			;W->FSR(pointer)
	movf	chrdata,W		;chrdata ->W
	movwf	INDF			;W -> *FSR
	xorlw	0x0d			;CR?
	btfsc	STATUS,Z
	goto	interpret		;yes,CR.

	incf	count,f			;count++
	movf	count,W
	sublw	0x08			;8-W
	btfsc	STATUS,C		;skip if C=0(borrow)
	goto	waitcmd			;C=1,not borrow, count=<8
;-------------------------------
interpret				;analyse command
	clrf	count

	call	GETCH			;W<=chrdata
	xorlw	'V'
	btfsc	STATUS,Z
	goto	voltset
	movf	chrdata,W
	xorlw	'R'
	btfsc	STATUS,Z
	goto	request
	goto	main
request
	incf	count,f			;++count
	call	GETCH			;W=chrdata
	xorlw	'S'
	btfsc	STATUS,Z
	goto	vreset
	movf	chrdata,W		;W=chrdata
	xorlw	'E'
	btfss	STATUS,Z
	goto	main
	incf	count,f
	call	GETCH
	xorlw	'Q'
	btfss	STATUS,Z
	goto	main
	incf	count,f
	call	GETCH
	xorlw	'A'
	btfsc	STATUS,Z
	goto	ackamp
	movf	chrdata,W
	xorlw	'P'
	btfsc	STATUS,Z
	goto	ackpnum
	movf	chrdata,W
	xorlw	'S'
	btfsc	STATUS,Z
	goto	acksern
	goto	main
;-------------------------------
vreset					;volt reset
	incf	count,f			;++count
	call	GETCH			;W=chrdata
	xorlw	'T'
	btfss	STATUS,Z
	goto	main
	clrf	pos_tgt			;reset output
	clrf	neg_tgt
	goto	main
;-------------------------------
ackpnum					;product number
	incf	count,f			;++count
	call	GETCH			;W=chrdata
	xorlw	'N'
	btfss	STATUS,Z
	goto	main
	clrf	count
mess1b	movf	count,W
	call	text_pnum
	xorlw	0x00
	btfsc	STATUS,Z
	goto	main
	movwf	chrdata
	call	TXBYTE
	incf	count,f
	goto	mess1b
;-------------------------------
acksern					;serial number
	incf	count,f			;++count
	call	GETCH			;W=chrdata
	xorlw	'N'
	btfss	STATUS,Z
	goto	main
	clrf	count
mess1c	movf	count,W
	call	text_snum
	xorlw	0x00
	btfsc	STATUS,Z
	goto	main
	movwf	chrdata
	call	TXBYTE
	incf	count,f
	goto	mess1c
;-------------------------------
ackamp					;current
	clrf	count
mess1a	movf	count,W
	call	text_acka
	xorlw	0x00
	btfsc	STATUS,Z
	goto	next1a
	movwf	chrdata
	call	TXBYTE
	incf	count,f
	goto	mess1a
next1a
	movlw	'0'
	movwf	chrdata
	call	TXBYTE

	movf	pos_curr,W
	movwf	_E
	call	bintod

	movlw	'0'
	movwf	chrdata
	call	TXBYTE

	movf	neg_curr,W
	movwf	_E
	call	bintod

	movlw	0x0d
	movwf	chrdata
	call	TXBYTE

	goto	main
;-------------------------------
voltset					;set positive voltage
	incf	count,f			;count++
	call	GETCH
	movlw	0x30			;W<-0x30
	subwf	chrdata,f		;chrdata-W ->chrdata
	btfss	STATUS,C
	goto	main			;<0
	movlw	0x02
	subwf	chrdata,W		;chrdata-2
	btfsc	STATUS,C
	goto	main
	clrf	_C
	movf	chrdata,W		;movwf	chrdata
	call	dtobin
	incf	count,f
	call	GETCH
	movlw	0x30
	subwf	chrdata,f
	btfss	STATUS,C
	goto	main
	movlw	0x0a			;W=10
	subwf	chrdata,W		;chrdata-W ->W
	btfsc	STATUS,C
	goto	main
	movf	chrdata,W		;movwf	chrdata
	call	dtobin
	incf	count,f
	call	GETCH
	movlw	0x30
	subwf	chrdata,f
	btfss	STATUS,C
	goto	main
	movlw	0x0a
	subwf	chrdata,W
	btfsc	STATUS,C
	goto	main
	movf	chrdata,W		;movwf	chrdata
	call	dtobin
	movwf	ptgtvolt		;W->ptgtvolt
	movlw	0x80			;W=0x80
	subwf	ptgtvolt,W		;ptgtvolt-W ->W
	btfss	STATUS,C
	goto	noadj			;C=0:borrow
	movlw	0x7f			;C=1:not borrow, ptgtvolt>=128
	movwf	ptgtvolt
noadj
	bcf	STATUS,C
	rlf	ptgtvolt,f
	movf	ptgtvolt,W
	movwf	pos_tgt

;-------------------------------	;set negative voltage
	incf	count,f			;count++
	call	GETCH
	movlw	0x30			;W<-0x30
	subwf	chrdata,f		;chrdata-W ->chrdata
	btfss	STATUS,C
	goto	main			;<0
	movlw	0x02
	subwf	chrdata,W		;chrdata-2
	btfsc	STATUS,C
	goto	main
	clrf	_C
	movf	chrdata,W		;movwf	chrdata
	call	dtobin
	incf	count,f
	call	GETCH
	movlw	0x30
	subwf	chrdata,f
	btfss	STATUS,C
	goto	main
	movlw	0x0a			;W=10
	subwf	chrdata,W		;chrdata-W ->W
	btfsc	STATUS,C
	goto	main
	movf	chrdata,W		;movwf	chrdata
	call	dtobin
	incf	count,f
	call	GETCH
	movlw	0x30
	subwf	chrdata,f
	btfss	STATUS,C
	goto	main
	movlw	0x0a
	subwf	chrdata,W
	btfsc	STATUS,C
	goto	main
	movf	chrdata,W		;movwf	chrdata
	call	dtobin
	movwf	ntgtvolt		;W->ntgtvolt
	movlw	0x80			;W=0x80
	subwf	ntgtvolt,W		;ntgtvolt-W ->W
	btfss	STATUS,C
	goto	nnoadj			;C=0:borrow
	movlw	0x7f			;C=1:not borrow, ptgtvolt>=128
	movwf	ntgtvolt
nnoadj
	bcf	STATUS,C
	rlf	ntgtvolt,f
	movf	ntgtvolt,W
	movwf	neg_tgt

	goto	main

;**************************** SUBROUTINE ******************************
TXBYTE:					;send byte to host
wloop	BTFSC	PORTC,4			;TXE#=L?
	GOTO	wloop			;NOP if L
	BSF	STATUS,RP0		;bank1
	MOVLW	0x00			;DIRportB=00000000
	MOVWF	TRISB			;0:out
	BCF	STATUS,RP0		;bank0
	BSF	PORTC,3			;WR->H
	MOVF	chrdata,W		;data->W
	MOVWF	PORTB
	BCF	PORTC,3			;WR->L
	BSF	STATUS,RP0		;bank1
	MOVLW	0xff			;DIRportB=11111111
	MOVWF	TRISB			;1:in
	BCF	STATUS,RP0		;bank0
	RETURN

GETCH:					;get character
	movf	count,W
	movlw	_data			;top address ->W
	addwf	count,W			;W + count ->W
	movwf	FSR			;W->FSR(pointer)
	movf	INDF,W			;*FSR ->W
	movwf	chrdata			;W ->chrdata
	return

dtobin:					;ASCII decimal -> binary
	movwf	_D			;save new, W->D
	movf	_C,W			;C->W
	addwf	_C,f			;C+W ->C :x2
	movf	_C,W			;C->W
	addwf	_C,W			;C+W ->W :x4
	movwf	_E			;W ->E
	addwf	_E,W			;E+W ->W :x8
	addwf	_C,W			;C+W ->W :x10
	addwf	_D,W			;D+W ->W :combine
	movwf	_C			;save in C
	return

bintod:					;binary -> ASCII decimal
	movlw	D'100'
	movwf	_D
dec81	movlw	'0'-1
	movwf	_C
dec82	incf	_C,f
	movf	_D,W
	subwf	_E,f			;100 or 10
	btfsc	STATUS,C
	goto	dec82			;still +
	addwf	_E,f			;add back
	movf	_C,W
	movwf	chrdata
	call	TXBYTE			;print it
	movlw	D'90'
	subwf	_D,f			;100 to 10
	btfsc	STATUS,C
	goto	dec81			;again
	movf	_E,W
	addlw	'0'			;ASCII bias
	movwf	chrdata
	call	TXBYTE
	return

	END                       	; directive 'end of program'

